// Copyright (C) 2002-2005 Nikolaus Gebhardt
// This file is part of the "Irrlicht Engine" and the "irrXML" project.
// For conditions of distribution and use, see copyright notice in irrlicht.h and/or irrXML.h

#ifndef __IRR_XML_H_INCLUDED__
#define __IRR_XML_H_INCLUDED__

#include "irrTypes.h"
	
	//! Enumeration of all supported source text file formats 
	enum ETEXT_FORMAT
	{
		//! ASCII, file without byte order mark, or not a text file
		ETF_ASCII,

		//! UTF-8 format
		ETF_UTF8,

		//! UTF-16 format, big endian
		ETF_UTF16_BE,

		//! UTF-16 format, little endian
		ETF_UTF16_LE,

		//! UTF-32 format, big endian
		ETF_UTF32_BE,

		//! UTF-32 format, little endian
		ETF_UTF32_LE,
	};


	//! Enumeration for all xml nodes which are parsed by IrrXMLReader
	enum EXML_NODE
	{
		//! No xml node. This is usually the node if you did not read anything yet.
		EXN_NONE,

		//! A xml element, like <foo>
		EXN_ELEMENT,

		//! End of an xml element, like </foo>
		EXN_ELEMENT_END,

		//! Text within a xml element: <foo> this is the text. </foo>
		EXN_TEXT,

		//! An xml comment like &lt;!-- I am a comment --&gt; or a DTD definition.
		EXN_COMMENT,

		//! An xml cdata section like &lt;![CDATA[ this is some CDATA ]]&gt;
		EXN_CDATA,

		//! Unknown element.
		EXN_UNKNOWN
	};
	

	//! Interface providing easy read access to a XML file.
	/** You can create an instance of this reader using one of the factory functions
	createIrrXMLReader(), createIrrXMLReaderUTF16() and createIrrXMLReaderUTF32().
	If using the parser from the Irrlicht Engine, please use IFileSystem::createXMLReader() 
	instead.
	For a detailed intro how to use the parser, see \ref irrxmlexample and \ref features.

	The typical usage of this parser looks like this:
	\code
	#include <irrXML.h>
	using namespace irr; // irrXML is located in the namespace irr::io
	using namespace io;

	void main()
	{
		// create the reader using one of the factory functions
		IrrXMLReader* xml = createIrrXMLReader("config.xml");

		if (xml == 0)
			return; // file could not be opened

		// parse the file until end reached
		while(xml->read())
		{
			// based on xml->getNodeType(), do something.
		}

		// delete the xml parser after usage
		delete xml;
	}
	\endcode
	See \ref irrxmlexample for a more detailed example.
	*/



	//! defines the utf-16 type.
	/** Not using wchar_t for this because 
	wchar_t has 16 bit on windows and 32 bit on other operating systems. */
	typedef unsigned short char16;

	//! defines the utf-32 type. 
	/** Not using wchar_t for this because 
	wchar_t has 16 bit on windows and 32 bit on other operating systems. */
	typedef unsigned long char32;

#include "irrXML.h"
#include "irrString.h"
#include "irrArray.h"
#include "fast_atof.h"
#include "CXMLReaderImpl.h"
#include <string.h>

/*
//! Implementation of the file read callback for ordinary files
class BufferReadCallBack {
public:

	//! construct from filename
	BufferReadCallBack(const c8* buffer, int size)
		: m_buffer(buffer), m_size(size), m_index(0)
	{
	}

	//! destructor
	~BufferReadCallBack()
	{
        m_buffer = NULL;
        m_size = 0;
        m_index = 0;
	}

	//! Reads an amount of bytes from the file.
	int read(void* buffer, int sizeToRead)
	{
        int amount = (sizeToRead > (m_size - m_index) ? (m_size - m_index) : sizeToRead);
        memcpy(buffer, m_buffer + m_index, amount);
        m_index += amount;
        return amount;
	}

	//! Returns size of file in bytes
	int getSize()
	{
		return m_size;
	}

private:

    const c8* m_buffer;
	int         m_size;
    int         m_index;

}; 
*/

	
//! Implementation of the file read callback for ordinary files
class FileReadCallBack {
public:

	FileReadCallBack(const char* fileName)
		: fileH(0), size(0), close(true)
	{
		Err error = errNone;
		fileH = FileOpen(0, fileName, 0, 'MIHA', fileModeReadOnly, &error);
		if(error != errNone){
			ErrAlert(error);
			error = errNone;
		} else {		
			getFileSize();
		}
	}
	
	~FileReadCallBack()
	{
		//if (close && fileH)
			FileClose(fileH);
	}
	
	//! Reads an amount of bytes from the file.
	int read(void* buffer, int sizeToRead)
	{
		if (!fileH)
			return 0;

		Err error = errNone;
		return (int)FileRead(fileH, buffer, 1, sizeToRead, &error);
	    if(error != errNone){
		    ErrAlert(error);
       		error = errNone;
	    }	    
	}

	//! Returns size of file in bytes
	int getSize()
	{
		return size;
	}

private:

	//! retrieves the file size of the open file
	void getFileSize()
	{
		Err error = errNone;
		FileSeek(fileH, 0, fileOriginEnd);
		FileTell(fileH, &size, &error);
		FileSeek(fileH, 0, fileOriginBeginning);
		
		if(error != errNone){
		    ErrAlert(error);
       		error = errNone;
	    }	    
	}
	
	
	FileHand fileH;
	Int32    size;
	bool     close;
}; 



#endif // __IRR_XML_H_INCLUDED__
